@@ -5,7 +5,7 @@ use POSIX qw(:errno_h);
 use Socket qw(PF_INET SOCK_STREAM pack_sockaddr_in inet_ntoa $CRLF SOL_SOCKET SO_ERROR);
 use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
 
-our $VERSION = "0.19";
+our $VERSION = "0.20";
 
 sub Hijk::Error::CONNECT_TIMEOUT         () { 1 << 0 } # 1
 sub Hijk::Error::READ_TIMEOUT            () { 1 << 1 } # 2
@@ -27,7 +27,7 @@ sub _read_http_message {
     my $header = $head_as_array ? [] : {};
     my $no_content_len = 0;
     my $head = "";
-    my $method_is_head = do { no warnings qw(uninitialized); $method eq "HEAD" };
+    my $method_has_no_content = do { no warnings qw(uninitialized); $method eq "HEAD" };
     my $close_connection;
     vec(my $rin = '', $fd, 1) = 1;
     do {
@@ -38,7 +38,7 @@ sub _read_http_message {
         return ($close_connection, $proto, $status_code, $header, $body)
             if $no_content_len && $decapitated && (!defined($nbytes) || $nbytes == 0);
         if (!defined($nbytes)) {
-            next if ($! == EWOULDBLOCK || $! == EAGAIN);
+            next if ($! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR);
             return (
                 $close_connection, undef, 0, undef, undef,
                 Hijk::Error::RESPONSE_READ_ERROR,
@@ -71,6 +71,7 @@ sub _read_http_message {
                 $head = substr($head, 0, $neck_pos);
                 $proto = substr($head, 0, 8);
                 $status_code = substr($head, 9, 3);
+                $method_has_no_content = 1 if $status_code == 204; # 204 NO CONTENT, see http://tools.ietf.org/html/rfc2616#page-60
                 substr($head, 0, index($head, $CRLF) + 2, ""); # 2 = length($CRLF)
 
                 my ($doing_chunked, $content_length, $trailer_mode, $trailer_value_is_true);
@@ -124,15 +125,19 @@ sub _read_http_message {
                     );
                 }
 
-                if ($content_length) {
-                    $read_length = $content_length - length($body);
+                if (defined $content_length) {
+                    if ($content_length == 0) {
+                        $read_length = 0;
+                    } else {
+                        $read_length = $content_length - length($body);
+                    }
                 } else {
                     $read_length = 10204;
                     $no_content_len = 1;
                 }
             }
         }
-    } while( !$decapitated || (!$method_is_head && ($read_length > 0 || $no_content_len)) );
+    } while( !$decapitated || (!$method_has_no_content && ($read_length > 0 || $no_content_len)) );
     return ($close_connection, $proto, $status_code, $header, $body);
 }
 
@@ -151,7 +156,7 @@ sub _read_chunked_body {
             my $current_buf = "";
             my $nbytes = POSIX::read($fd, $current_buf, $read_length);
             if (!defined($nbytes)) {
-                next if ($! == EWOULDBLOCK || $! == EAGAIN);
+                next if ($! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR);
                 return (
                     undef,
                     Hijk::Error::RESPONSE_READ_ERROR,
@@ -332,7 +337,7 @@ sub request {
 
         my $rc = syswrite($soc,$r,$left, $total - $left);
         if (!defined($rc)) {
-            next if ($! == EWOULDBLOCK || $! == EAGAIN);
+            next if ($! == EWOULDBLOCK || $! == EAGAIN || $! == EINTR);
             delete $args->{socket_cache}->{$cache_key} if defined $cache_key;
             shutdown($soc, 2);
             return {
